
public class MSAMotif {
	
	public final static int INVALID_ALIGNMENT = -10; 	 // flag used to mark badly aligned k-mers 
	
	private int topT;							 // number of top k-mers requested for alignment
	private int numAligned;						 // number of k-mers actually aligned, after filtering 
	private String center;						 // the center string of the alignment
	private MerObj kMers[];						 // the k-mers array
	private Integer[] kMersSort;				 // the k-mers ranks array
	private int alignmentLength;				 // the length of the alignment
	private int maxLeftOffset;					 // max left offset of a k-mer vis-a-vis the center
	private int maxRightOffset;					 // max right offset of a k-mer vis-a-vis the center							
	private float[][] maxPositionMatrix;	 	 // PWM covering max number of top k-mers
	private float[][] midPositionMatrix;		 // PWM covering mid number of top k-mers
	private float[][] minPositionMatrix;		 // PWM covering min number of top k-mers
	private float[][] finalPWM;					 // weighted mix of above matrices
	private int MAX_OFFSET;
	private float MIN_MATCHES;
	
	public MSAMotif (MerObj[] kMers, Integer[] kMerRanks, int maxOffset, float minMatches) {
		this.topT = Application.MAX_KMERS;
		this.numAligned = Application.MAX_KMERS;
		this.center = kMers[kMerRanks[0]].getStr(); // the center string is the top ranking one
		this.kMers = kMers;
		this.kMersSort = kMerRanks;
		this.alignmentLength = this.center.length();
		this.maxLeftOffset = 0;
		this.maxRightOffset = 0;
		this.MAX_OFFSET = maxOffset;
		this.MIN_MATCHES = minMatches;
		
		this.alignAll();
	}
	
	/* Align all k-mers to the center string with their best offsets */
	private void alignAll() {
		this.kMers[this.kMersSort[0]].setAlignOffset(0);
		for (int i = 1; i < this.topT; i++) {
			int bestOffset;
			bestOffset = align(this.kMers[this.kMersSort[i]]);
			this.kMers[this.kMersSort[i]].setAlignOffset(bestOffset);  // maintain best offset of each k-mer 
			if(bestOffset > 0 && bestOffset != INVALID_ALIGNMENT && bestOffset > maxRightOffset) {
				this.maxRightOffset = bestOffset; // update max right offset field
			}
			if(bestOffset < 0 && bestOffset != INVALID_ALIGNMENT && bestOffset < maxLeftOffset) {
				this.maxLeftOffset = bestOffset;  // update max left offset field
			}
		}
		this.alignmentLength += (maxRightOffset + (-1 * maxLeftOffset)); // compute alignment length
	}
	
	/* Align one k-mer against the center string and return its best-matched offset */
	private int align(MerObj kMer) {
		float maxMatchesWatson = 0; // keeps the max matches found on one strand
		float maxMatchesCrick = 0;	// keeps the max matches found on the other strand
		int bestWatsonOffset = 0;	// same for best offset from center
		int bestCrickOffset = 0;	// same for best offset from center 
		kMer.setRc(reverseComplement(kMer.getStr())); // set reverse complement of the k-mer
		String str = kMer.getStr();
		String rcStr = kMer.getRcStr();
		
		for (int i = -1 * MAX_OFFSET; i <= MAX_OFFSET; i++) { // try each allowable offset in turn
			float matchesWatson = checkMatches(this.center, str, i); // get #matches with offset for each orientation
			float matchesCrick = checkMatches(this.center, rcStr, i);
			if(matchesWatson > maxMatchesWatson) { // keep best matched offset for each orientation 
				maxMatchesWatson = matchesWatson;
				bestWatsonOffset = i;
			}
			if(matchesCrick > maxMatchesCrick) {
				maxMatchesCrick = matchesCrick;
				bestCrickOffset = i;
			}
		}
		
		if(maxMatchesWatson >= maxMatchesCrick) { // determine which orientation is better matched
			if(maxMatchesWatson >= MIN_MATCHES) {return bestWatsonOffset;} // make sure best match is significant enough  
			else { // otherwise, discard
				this.numAligned--;
				return INVALID_ALIGNMENT;
			}
		} else {
			kMer.setRcFlag();					 // mark this k-mer for best match in the rc
			if(maxMatchesCrick >= MIN_MATCHES) {return bestCrickOffset;} 
			else {
				this.numAligned--;
				return INVALID_ALIGNMENT;
			}
		}		
	}
	
	/* Call helper methods to compute the PWM, then determine significant positions and return final
	   motif PWM to caller 
	*/
	public double[][] getMotifPWM(char[][] probes, float[] probeScores) {
		computeAuxilliaryPWMs(probes, probeScores);
		computeFinalPWM();
		
		int leftSignificanceBoundary = -1;
		int rightSignificanceBoundary = -1;
		for (int i = 0; i < Application.MAX_MOTIF; i++) { // mark significance boundaries
			if(entropyTest(this.finalPWM[i])) {
				rightSignificanceBoundary = i;
				if(leftSignificanceBoundary == -1) { leftSignificanceBoundary = i; }
			}	
		}
		
		double[][] PWM = new double[rightSignificanceBoundary - leftSignificanceBoundary + 1][4];
		int m = 0;
		int p = 0;
		for (int i = leftSignificanceBoundary; i <= rightSignificanceBoundary; i++) {
			p = 0;
			for (int j = 0; j < 4; j++) {
				PWM[m][p] = (double)this.finalPWM[i][j];
				p++;
			}
			m++;
		}
		
		return PWM;
	}
	
	/* Populate 3 PWM matrices, each of the size of the max motif, for each layer of the 
	   top k-mers (max, mid and min as defined in the Application). 
	   The distributions in each position are weighted by the average binding scores of the k-mers 
	   considered.
	*/
	private void computeAuxilliaryPWMs(char[][] probes, float[] probeScores) {
		this.maxPositionMatrix = new float[Application.MAX_MOTIF][4];
		this.midPositionMatrix = new float[Application.MAX_MOTIF][4];
		this.minPositionMatrix = new float[Application.MAX_MOTIF][4];
		
		/* Variables for holding the sum of avg. binding scores for each matrix, for normalization purposes */
		float maxSumOfAverageScores = 0;
		float midSumOfAverageScores = 0; 
		float minSumOfAverageScores = 0; 
		int firstPositionLoop = 1; // this flag will make sure kmers add their avg binding score to the sum only once
		
		int motifMargin = (Application.MAX_MOTIF - Application.MER_SIZE) / 2; // number of extra positions to compute, bringing the total to MAX_MOTIF
		int posOffset = -motifMargin;
		int column = -1;
		while(posOffset < Application.MER_SIZE + motifMargin) { // loop over all positions
			int cutoffCounter = 0;
			int j = 0;
			while (cutoffCounter < this.numAligned) {			// loop over aligned k-mers
				MerObj kMer = this.kMers[this.kMersSort[j]];
				int alignOffset = kMer.getAlignOffset();
				if(alignOffset == INVALID_ALIGNMENT) {			// skip over badly aligned k-mers
					j++;
					continue;
				} else {
					j++;
					cutoffCounter++;
					int rcFlag = kMer.getRcFlag();
					float enteredProbes = 0;
					int k = 0;
					int numTotalProbes = kMer.getForegroundSize();
					while (enteredProbes < Application.PROBE_CUTOFF && k < numTotalProbes) {	// loop over k-mer probes
						int rank = kMer.getForegroundRank(k);
						int offsetInProbe = kMer.getProbeOffset(k);
						int nextPos;
						char c;
						if(rcFlag == 0) {	// get the position on the probe if this is the + strand
							nextPos = offsetInProbe - alignOffset + posOffset;
						} else {  			// same for the - strand
							nextPos = offsetInProbe + (Application.MER_SIZE-1) + alignOffset - posOffset;  
						} 
						if(nextPos >= Application.PROBE_SIZE || nextPos < 0) {	// if next position out of probe bounds, skip it
							k++;
							continue;
						}
						enteredProbes++; k++;
						if(rcFlag == 0) {	
							c = probes[rank-1][nextPos];	// get nucleotide at position
						} else {
							c = complement(probes[rank-1][nextPos]);	// same for rc
						}				
						switch (c) {
							case ('A'): { column = 0; break; }
							case ('C'): { column = 1; break; }
							case ('G'): { column = 2; break; }
							case ('T'): { column = 3; break; }
							default: { break; }
						}
						this.maxPositionMatrix[posOffset+motifMargin][column] += kMer.getAvg();
						if(cutoffCounter < Application.MID_KMERS) {
							this.midPositionMatrix[posOffset+motifMargin][column] += kMer.getAvg();
						}
						if(cutoffCounter < Application.MIN_KMERS) {
							this.minPositionMatrix[posOffset+motifMargin][column] += kMer.getAvg();
						}
					}
					if(firstPositionLoop == 1) { maxSumOfAverageScores += kMer.getAvg(); }
					if(firstPositionLoop == 1 && cutoffCounter < Application.MID_KMERS) { midSumOfAverageScores += kMer.getAvg(); }
					if(firstPositionLoop == 1 && cutoffCounter < Application.MIN_KMERS) { minSumOfAverageScores += kMer.getAvg(); }	
				}
			}
			firstPositionLoop = 0;
			for (int k = 0; k < 4; k++) {
				this.maxPositionMatrix[posOffset+motifMargin][k] = (float) (this.maxPositionMatrix[posOffset+motifMargin][k] / (maxSumOfAverageScores * (float)Application.PROBE_CUTOFF) );
				this.midPositionMatrix[posOffset+motifMargin][k] = (float) (this.midPositionMatrix[posOffset+motifMargin][k] / (midSumOfAverageScores * (float)Application.PROBE_CUTOFF) );
				this.minPositionMatrix[posOffset+motifMargin][k] = (float) (this.minPositionMatrix[posOffset+motifMargin][k] / (minSumOfAverageScores * (float)Application.PROBE_CUTOFF) );
			}
			posOffset++;
		}
		return;
	}
	
	private void computeFinalPWM() {
		this.finalPWM = new float[Application.MAX_MOTIF][4];
		
		int[] posDesignation = designateCorePositions(); // mark positions conserved at >=CORE_THRESHOLD in the max PWM
		int leftCoreBoundary = -1;
		int rightCoreBoundary = -1;
		for (int i = 0; i < Application.MAX_MOTIF; i++) { // mark core boundaries
			if(posDesignation[i] == 1) {
				rightCoreBoundary = i;
				if(leftCoreBoundary == -1) { leftCoreBoundary = i; }
			}	
		}
		
		for (int i = 0; i < Application.MAX_MOTIF; i++) { // mark positions by distance from core
			if(posDesignation[i] == 1) { posDesignation[i] = 0; } 
			else if (i < leftCoreBoundary) {										
				posDesignation[i] = leftCoreBoundary - i;
			} else if (i > rightCoreBoundary) {
				posDesignation[i] = i - rightCoreBoundary;
			} else {
				posDesignation[i] = 1;
			}
		}

		for (int i = 0; i < Application.MAX_MOTIF; i++) { // mix final PWM according to position
			for (int j = 0; j < 4; j++) {
				finalPWM[i][j] = mixPWMs(i, j, Math.min(2, posDesignation[i]));
			}
		}
		
		return;
	}
	
	/* Mark which positions in the maximal PWM matrix is conserved above the threshold */
	private int[] designateCorePositions() {
		int[] designation = new int[Application.MAX_MOTIF];
		for(int i = 0; i < Application.MAX_MOTIF; i++) {
			for(int j = 0; j < 4; j++) {
				if(this.maxPositionMatrix[i][j] >= Application.CORE_THRESHOLD) {
					designation[i] = 1;
					break;
				}
			}
		}
		return designation;
	}
	
	/* Mix the auxiliary PWMs based on the necessary formula */
	private float mixPWMs(int i, int j, int policy) {
		float maxFrac = 0;
		float midFrac = 0;
		float minFrac = 0;
		switch(policy) {
			case(0): maxFrac = Application.MAX_FIRST_FRAC; midFrac = Application.MID_FIRST_FRAC; minFrac = Application.MIN_FIRST_FRAC; break;
			case(1): maxFrac = Application.MAX_SECOND_FRAC; midFrac = Application.MID_SECOND_FRAC; minFrac = Application.MIN_SECOND_FRAC; break;
			case(2): maxFrac = Application.MAX_THIRD_FRAC; midFrac = Application.MID_THIRD_FRAC; minFrac = Application.MIN_THIRD_FRAC; break;
		}
		return (this.maxPositionMatrix[i][j] * maxFrac + this.midPositionMatrix[i][j] * midFrac + this.minPositionMatrix[i][j] * minFrac);  
	}
	
	/* Compute entropy measure of a position and determine if it is significant */
	private boolean entropyTest(float[] position) {
		float entropy = 0;
		for (int i = 0; i < 4; i++) {
			if(position[i] == (float)0) {continue;}
			entropy += (float) (position[i] * Math.log10(position[i]) / Math.log10(2));
		}
		if(-1*entropy < (float)Application.ENTROPY_THRESHOLD) { return true; }
		else { return false; }
	}
	
/* ======================================UTILITY FUNCTIONS===================================== */
	
	private String reverseComplement(String s) {
		StringBuilder rc = new StringBuilder(Application.MER_SIZE);
		for (int i = s.length() - 1; i > -1; i--) {
			rc.append(complement(s.charAt(i)));
		}
		return rc.toString();
	}
	
	private char complement(char c) {
		switch (c) {
			case ('A'): { return 'T'; }
			case ('C'): { return 'G'; }
			case ('G'): { return 'C'; }
			case ('T'): { return 'A'; }
		}
		return ('X');
	}

	/* Count and return #matches between two strings at the specified offset, adding 0.1 for each
	   consecutive match to give preference to consecutive over scattered matches 
	*/
	private float checkMatches(String s1, String s2, int offset) {
		int matchCounter = 0;
		int consecMatchCounter = 0; 
		float maxConsec = 0; // also remember the maximum number of consecutive matches 
		
		int s1Index;
		int s2Index;
		if(offset >= 0) {
			s1Index = offset;
			s2Index = 0;
		} else {
			s1Index = 0;
			s2Index = -1 * offset;
		}
		while(s1Index < s1.length() && s2Index < s2.length()) {
			if(s1.charAt(s1Index) == s2.charAt(s2Index)) {
				matchCounter++;
				consecMatchCounter++;
			} else {
				if(consecMatchCounter > maxConsec) {maxConsec = consecMatchCounter;}
				consecMatchCounter = 0;
			}
			s1Index++;
			s2Index++;
		}

		return matchCounter + maxConsec/(float)10.0; 	
	}
}
